import { nanoid } from 'nanoid';

var Configs = {
  idPrefix: "w",
  rootElement: "[data-widgets-root]",
  widgetSelector: {
    datasetHtmlAttribute: "data-widget",
    // for querySelector('[data-widget]')
    datasetKey: "widget" // for dataset.widget

  },
  widgetId: {
    datasetHtmlAttribute: "data-widget-id",
    // for querySelector('[data-widget-id]')
    datasetKey: "widgetId" // for dataset.widgetId

  },
  htmlProps: {
    prefix: ":"
  }
};

const convertType = v => {
  const value = v.trim();

  if (value.startsWith("{") && value.endsWith("}") || value.startsWith("[") && value.endsWith("]")) {
    return JSON.parse(value.replace(/'/g, '"'));
  } else {
    switch (value) {
      case "false":
        return false;

      case "true":
        return true;

      case "null":
        return null;

      default:
        return !isNaN(value) ? parseFloat(value) : value;
    }
  }
};
const createWidget = ({
  name,
  template,
  tag,
  props,
  className
}) => {
  const rootTag = document.createElement(tag || "div");
  if (name) rootTag.dataset[Configs.widgetSelector.datasetKey] = name;

  if (className) {
    rootTag.classList.add(className);
  }

  if (template) {
    rootTag.innerHTML = template;
  }

  if (props) {
    for (const [key, value] of Object.entries(props)) {
      rootTag.setAttribute(`${Configs.htmlProps.prefix}${key}`, value);
    }
  }

  return rootTag;
};

class WidgetInstance {
  constructor(htmlEl, handler, plugins) {
    this.id = `${Configs.idPrefix}${nanoid(6)}`;
    this.$el = htmlEl;
    const p = this.collectProps();
    this.props = p;
    this.plugins = plugins;
    this.widgetHandlerFunction = handler;
    this.observables = {};
    this.init();
  }

  propEffect = (propName, effect) => {
    const obj = {
      current: this.props[propName]
    };
    const handler = {
      get(obj, prop) {
        return obj[prop];
      },

      set(obj, prop, value) {
        obj[prop] = value;
        effect();
        return true;
      }

    };
    const proxy = new Proxy(obj, handler);
    this.observables[propName] = proxy;
    return [this.observables[propName], newValue => {
      this.$el.setAttribute(`:${String(propName)}`, newValue);
    }];
  };

  watchAttributes() {
    const observer = new MutationObserver(mutationsList => {
      for (let mutation of mutationsList) {
        if (mutation.type !== "attributes") return;
        const prop = mutation.attributeName.substring(1);
        const newValue = convertType(this.$el.getAttribute(mutation.attributeName));
        this.observables[prop].current = newValue;
      }
    });
    observer.observe(this.$el, {
      attributes: true,
      childList: false,
      subtree: false
    });
  }

  get CONTEXT() {
    return {
      id: this.id,
      $el: this.$el,
      props: this.props,
      propEffect: this.propEffect
    };
  }

  get PLUGINS() {
    return this.plugins(this.CONTEXT);
  } // add the id as data attribute to the current HTMLElement


  setWidgetIdAttribute() {
    const {
      $el,
      id
    } = this;
    $el.setAttribute(Configs.widgetId.datasetHtmlAttribute, id);
  }

  collectProps = () => {
    const obj = {};
    Object.values(this.$el.attributes).filter(el => el.name.startsWith(Configs.htmlProps.prefix)).forEach(el => {
      obj[el.name.substring(1)] = convertType(el.value);
    });
    return obj;
  };

  init() {
    // add the id as data attribute, it will be used to run the destroy method
    this.setWidgetIdAttribute();
    this.watchAttributes(); // call widget function

    const widgetHandler = this.widgetHandlerFunction(this.CONTEXT, this.PLUGINS); // if the widget return a function use it as destroyer handler

    this.destroy = widgetHandler ? widgetHandler : () => {};
  }

}

const RH_MEMORY = new Map();
class Observer {
  constructor({
    lazyImport,
    plugins,
    widgets,
    asyncWidgets,
    rootElement = Configs.rootElement,
    selector = `[${Configs.widgetSelector.datasetHtmlAttribute}]`,
    logs = false
  }) {
    this.plugins = plugins;
    this.lazyImport = lazyImport;
    this.shouldLog = logs;
    this.rootElement = document.body.querySelector(rootElement);
    this.selector = selector;
    this.widgets = widgets;
    this.asyncWidgets = asyncWidgets;
    this.init();
  }

  get COMPONENT_LIST() {
    return Object.keys(this.widgets || {});
  }

  afterNodeDeleted(removedNodes) {
    removedNodes.filter(el => el.dataset && el.dataset[Configs.widgetId.datasetKey]).forEach(comp => {
      const widgetId = comp.dataset[Configs.widgetId.datasetKey];
      const widgetName = comp.dataset[Configs.widgetSelector.datasetKey];

      if (RH_MEMORY.has(widgetId)) {
        const instance = RH_MEMORY.get(widgetId);

        if (this.shouldLog) {
          console.log(`%c[${widgetName} # ${widgetId}] destroyed`, "color: white; background-color: #9c27b0; padding: 3px 5px;");
          console.log(comp);
        }

        instance.destroy();
        RH_MEMORY.delete(widgetId);
      }
    });
  }

  afterNodeAdded(addedNodes) {
    addedNodes.filter(el => !!el.querySelectorAll).forEach(addedNode => {
      this.importWidgets(addedNode);
    });
  }

  async observeDomChanges(target) {
    const observer = new MutationObserver(mutationsList => {
      for (let mutation of mutationsList) {
        // handle only added and removed nodes
        if (mutation.type !== "childList") return;
        const addedNodes = Array.from(mutation.addedNodes);
        const removedNodes = Array.from(mutation.removedNodes);

        if (mutation.target && removedNodes.length) {
          this.afterNodeDeleted(removedNodes);
        }

        if (mutation.target && addedNodes.length) {
          this.afterNodeAdded(addedNodes);
        }
      }
    });
    observer.observe(target, {
      attributes: false,
      childList: true,
      subtree: true
    });
  }

  findWidgets(target) {
    const finalTarget = target !== this.rootElement ? target.parentNode : this.rootElement;
    return Array.from(finalTarget.querySelectorAll(this.selector)).filter(el => !!el.dataset[Configs.widgetSelector.datasetKey] && !el.dataset[Configs.widgetId.datasetKey]);
  }

  importWidgets(target) {
    return new Promise((resolve, reject) => {
      try {
        const widgets = this.findWidgets(target);
        widgets.forEach(async component => {
          const componentName = component.dataset[Configs.widgetSelector.datasetKey]; // if component is typeof string is considered a path to lazy import

          const shouldImport = !this.COMPONENT_LIST.includes(componentName);
          const p = this.plugins;
          let instance = null;

          if (shouldImport) {
            const asyncWidgetHandler = await this.lazyImport(componentName);
            instance = new WidgetInstance(component, asyncWidgetHandler.default, this.plugins);
          } else {
            const widgetHandler = this.widgets[componentName];
            instance = new WidgetInstance(component, widgetHandler, this.plugins);
          } // store component reference


          RH_MEMORY.set(instance.id, instance);

          if (this.shouldLog) {
            console.log(`%c${shouldImport ? "⚡️ " : ""}[${componentName}] initiated`, "color: white; background-color: #3f51b5; padding: 3px 5px;");
            console.log(instance.$el);
          }
        });
        resolve();
      } catch (e) {
        reject(e);
      }
    });
  }

  async init() {
    try {
      await this.importWidgets(this.rootElement);
      await this.observeDomChanges(this.rootElement);
    } catch (e) {
      console.error("WIDGETS-ERR", e);
    }
  }

}

export { createWidget, Observer as default };
